/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 2004
 *	Sleepycat Software.  All rights reserved.
 *
 * $Id: rep_log.c,v 12.11 2005/01/19 13:53:00 bostic Exp $
 */

#include "db_config.h"

#ifndef NO_SYSTEM_INCLUDES
#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#include <time.h>
#endif
#endif

#include <stdlib.h>
#include <string.h>
#endif

#include "db_int.h"
#include "dbinc/db_page.h"
#include "dbinc/db_am.h"
#include "dbinc/log.h"

/*
 * __rep_allreq --
 *      Handle a REP_ALL_REQ message.
 *
 * PUBLIC: int __rep_allreq __P((DB_ENV *, REP_CONTROL *, int));
 */
int
__rep_allreq(dbenv, rp, eid)
        DB_ENV *dbenv;
        REP_CONTROL *rp;
        int eid;
{
	DB_LOGC *logc;
	DB_LSN oldfilelsn;
	DB_REP *db_rep;
	DBT data_dbt;
	REP *rep;
	REP_THROTTLE repth;
	u_int32_t flags;
	int ret, t_ret;

	ret = 0;
	db_rep = dbenv->rep_handle;
	rep = db_rep->region;

	if ((ret = __log_cursor(dbenv, &logc)) != 0)
		return (ret);
	memset(&data_dbt, 0, sizeof(data_dbt));
	memset(&repth, 0, sizeof(repth));
	MUTEX_LOCK(dbenv, db_rep->rep_mutexp);
	repth.gbytes = rep->gbytes;
	repth.bytes = rep->bytes;
	oldfilelsn = repth.lsn = rp->lsn;
	repth.type = REP_LOG;
	repth.data_dbt = &data_dbt;
	MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
	flags = IS_ZERO_LSN(rp->lsn) ||
	    IS_INIT_LSN(rp->lsn) ?  DB_FIRST : DB_SET;
	for (ret = __log_c_get(logc, &repth.lsn, &data_dbt, flags);
	    ret == 0 && repth.type == REP_LOG;
	    ret = __log_c_get(logc, &repth.lsn, &data_dbt, DB_NEXT)) {
		if (repth.lsn.file != oldfilelsn.file)
			(void)__rep_send_message(dbenv,
			    eid, REP_NEWFILE, &oldfilelsn, NULL, 0);
		if (__rep_send_throttle(dbenv, eid, &repth) != 0)
			break;
		/*
		 * If we are about to change files, then we'll need the
		 * last LSN in the previous file.  Save it here.
		 */
		oldfilelsn = repth.lsn;
		oldfilelsn.offset += logc->c_len;
	}

	if (ret == DB_NOTFOUND)
		ret = 0;
	if ((t_ret = __log_c_close(logc)) != 0 && ret == 0)
		ret = t_ret;
	return (ret);
}

/*
 * __rep_log --
 *      Handle a REP_LOG/REP_LOG_MORE message.
 *
 * PUBLIC: int __rep_log __P((DB_ENV *, REP_CONTROL *, DBT *, 
 * PUBLIC:     time_t, DB_LSN *));
 */
int
__rep_log(dbenv, rp, rec, savetime, ret_lsnp)
        DB_ENV *dbenv;
        REP_CONTROL *rp;
        DBT *rec;
        time_t savetime;
	DB_LSN *ret_lsnp;
{
	DB_LOG *dblp;
	DB_LSN lsn;
	DB_REP *db_rep;
	LOG *lp;
	REP *rep;
	int is_dup, master, ret;

	is_dup = ret = 0;
	db_rep = dbenv->rep_handle;
	rep = db_rep->region;
	dblp = dbenv->lg_handle;
	lp = dblp->reginfo.primary;

	ret = __rep_apply(dbenv, rp, rec, ret_lsnp, &is_dup);
	switch (ret) {
	/*
	 * We're in an internal backup and we've gotten 
	 * all the log we need to run recovery.  Do so now.
	 */
	case DB_REP_LOGREADY:
		if ((ret = __log_flush(dbenv, NULL)) != 0)
			goto out;
		if ((ret = __rep_verify_match(dbenv, &rep->last_lsn,
		    savetime)) == 0) {
			MUTEX_LOCK(dbenv, db_rep->rep_mutexp);
			ZERO_LSN(rep->first_lsn);
			ZERO_LSN(rep->last_lsn);
			F_CLR(rep, REP_F_RECOVER_LOG);
			MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
		}
		break;
	/*
	 * If we get any of the "normal" returns, we only process
	 * LOG_MORE if this is not a duplicate record.  If the 
	 * record is a duplicate we don't want to handle LOG_MORE
	 * and request a multiple data stream (or trigger internal
	 * initialization) since this could be a very old record
	 * that no longer exists on the master.
	 */
	case DB_REP_ISPERM:
	case DB_REP_NOTPERM:
	case 0:
		if (is_dup)
			goto out;
		else
			break;
	/*
	 * Any other return (errors), we're done.
	 */
	default:
		goto out;
	}
	if (rp->rectype == REP_LOG_MORE) {
		MUTEX_LOCK(dbenv, db_rep->rep_mutexp);
		master = rep->master_id;
		MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
		R_LOCK(dbenv, &dblp->reginfo);
		lsn = lp->lsn;
		R_UNLOCK(dbenv, &dblp->reginfo);
		/*
		 * If the master_id is invalid, this means that since
		 * the last record was sent, somebody declared an
		 * election and we may not have a master to request
		 * things of.
		 *
		 * This is not an error;  when we find a new master,
		 * we'll re-negotiate where the end of the log is and
		 * try to bring ourselves up to date again anyway.
		 */
		MUTEX_LOCK(dbenv, db_rep->db_mutexp);
		if (master == DB_EID_INVALID)
			ret = 0;
		/*
		 * If we've asked for a bunch of records, it could
		 * either be from a LOG_REQ or ALL_REQ.  If we're
		 * waiting for a gap to be filled, call loggap_req,
		 * otherwise use ALL_REQ again.
		 */
		else if (IS_ZERO_LSN(lp->waiting_lsn)) {
			MUTEX_UNLOCK(dbenv, db_rep->db_mutexp);
			if (__rep_send_message(dbenv,
			    master, REP_ALL_REQ, &lsn, NULL, 0) != 0)
				goto out;
		} else {
			ret = __rep_loggap_req(dbenv, rep, &lsn, 1);
			MUTEX_UNLOCK(dbenv, db_rep->db_mutexp);
		}
	}
out:
	return (ret);
}


/*
 * __rep_log_req --
 *      Handle a REP_LOG_REQ message.
 *
 * PUBLIC: int __rep_logreq __P((DB_ENV *, REP_CONTROL *, DBT *, int));
 */
int
__rep_logreq(dbenv, rp, rec, eid)
        DB_ENV *dbenv;
        REP_CONTROL *rp;
        DBT *rec;
        int eid;
{
	DB_LOG *dblp;
	DB_LOGC *logc;
	DB_LSN endlsn, lsn, oldfilelsn;
	DB_REP *db_rep;
	DBT data_dbt;
	LOG *lp;
	REP *rep;
	REP_THROTTLE repth;
	int ret, t_ret;
#ifdef DIAGNOSTIC
	DB_MSGBUF mb;
#endif

	ret = 0;
	db_rep = dbenv->rep_handle;
	rep = db_rep->region;
	dblp = dbenv->lg_handle;
	lp = dblp->reginfo.primary;

	if (rec != NULL && rec->size != 0) {
		RPRINT(dbenv, rep, (dbenv, &mb,
		    "[%lu][%lu]: LOG_REQ max lsn: [%lu][%lu]",
		    (u_long) rp->lsn.file, (u_long)rp->lsn.offset,
		    (u_long)((DB_LSN *)rec->data)->file,
		    (u_long)((DB_LSN *)rec->data)->offset));
	}
	/*
	 * There are three different cases here.
	 * 1. We asked for a particular LSN and got it.
	 * 2. We asked for an LSN and it's not found because it is
	 *	beyond the end of a log file and we need a NEWFILE msg.
	 *	and then the record that was requested.
	 * 3. We asked for an LSN and it simply doesn't exist, but
	 *    doesn't meet any of those other criteria, in which case
	 *    it's an error (that should never happen).
	 * If we have a valid LSN and the request has a data_dbt with
	 * it, then we need to send all records up to the LSN in the
	 * data dbt.
	 */
	memset(&data_dbt, 0, sizeof(data_dbt));
	oldfilelsn = lsn = rp->lsn;
	if ((ret = __log_cursor(dbenv, &logc)) != 0)
		return (ret);
	ret = __log_c_get(logc, &lsn, &data_dbt, DB_SET);

	if (ret == 0) /* Case 1 */
		(void)__rep_send_message(dbenv,
		   eid, REP_LOG, &lsn, &data_dbt, DB_LOG_RESEND);
	else if (ret == DB_NOTFOUND) {
		R_LOCK(dbenv, &dblp->reginfo);
		endlsn = lp->lsn;
		R_UNLOCK(dbenv, &dblp->reginfo);
		if (endlsn.file > lsn.file) {
			/*
			 * Case 2:
			 * Need to find the LSN of the last record in
			 * file lsn.file so that we can send it with
			 * the NEWFILE call.  In order to do that, we
			 * need to try to get {lsn.file + 1, 0} and
			 * then backup.
			 */
			endlsn.file = lsn.file + 1;
			endlsn.offset = 0;
			if ((ret = __log_c_get(logc,
			    &endlsn, &data_dbt, DB_SET)) != 0 ||
			    (ret = __log_c_get(logc,
				&endlsn, &data_dbt, DB_PREV)) != 0) {
				RPRINT(dbenv, rep, (dbenv, &mb,
				    "Unable to get prev of [%lu][%lu]",
				    (u_long)lsn.file,
				    (u_long)lsn.offset));
				/*
				 * We want to push the error back
				 * to the client so that the client
				 * does an internal backup.  The
				 * client asked for a log record
				 * we no longer have and it is
				 * outdated.
				 * XXX - This could be optimized by
				 * having the master perform and
				 * send a REP_UPDATE message.  We
				 * currently want the client to set
				 * up its 'update' state prior to
				 * requesting REP_UPDATE_REQ.
				 */
				ret = 0;
				(void)__rep_send_message(dbenv, eid,
				    REP_VERIFY_FAIL, &rp->lsn, NULL, 0);
			} else {
				endlsn.offset += logc->c_len;
				(void)__rep_send_message(dbenv, eid,
				    REP_NEWFILE, &endlsn, NULL, 0);
			}
		} else {
			/* Case 3 */
			__db_err(dbenv,
			    "Request for LSN [%lu][%lu] fails",
			    (u_long)lsn.file, (u_long)lsn.offset);
			DB_ASSERT(0);
			ret = EINVAL;
		}
	}

	/*
	 * If the user requested a gap, send the whole thing,
	 * while observing the limits from set_rep_limit.
	 */
	memset(&repth, 0, sizeof(repth));
	MUTEX_LOCK(dbenv, db_rep->rep_mutexp);
	repth.gbytes = rep->gbytes;
	repth.bytes = rep->bytes;
	repth.lsn = oldfilelsn;
	repth.type = REP_LOG;
	repth.data_dbt = &data_dbt;
	MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
	while (ret == 0 && rec != NULL && rec->size != 0 &&
	    repth.type == REP_LOG) {
		if ((ret =
		    __log_c_get(logc, &repth.lsn, &data_dbt, DB_NEXT)) != 0) {
			if (ret == DB_NOTFOUND)
				ret = 0;
			break;
		}
		if (log_compare(&repth.lsn, (DB_LSN *)rec->data) >= 0)
			break;
		if (repth.lsn.file != oldfilelsn.file)
			(void)__rep_send_message(dbenv,
			    eid, REP_NEWFILE, &oldfilelsn, NULL, 0);
		if (__rep_send_throttle(dbenv, eid, &repth) != 0)
			break;
		/*
		 * If we are about to change files, then we'll need the
		 * last LSN in the previous file.  Save it here.
		 */
		oldfilelsn = repth.lsn;
		oldfilelsn.offset += logc->c_len;
	}

	if ((t_ret = __log_c_close(logc)) != 0 && ret == 0)
		ret = t_ret;
	return (ret);
}
